(*******************************************)
(*  Delphi 4.0 Simulation Components       *)
(*     Copyright (c) 2000 IzhSTU           *)
(*          Lusnikov Roman                 *)
(*******************************************)

unit DSComp;

interface

uses classes, sysutils, controls, graphics, windows,
     messages;

type
  TDSTime=integer;//cardinal;

  EDSError=class(Exception);
  TDSDevice=class;
  TDSTimer=class;

  TTransact=class
  private
    FTag: integer;
    FPriority: integer;          //
    FPower: integer;             //   
    FTime: TDSTime;              //   
    FParam: Pointer;             //  
    FParent: TDSDevice;          //,      
    procedure Assign(Source: TTransact);
  public
    constructor Create;
    property Priority:integer read FPriority write FPriority;
    property Tag:integer read FTag write FTag;
    property Power:integer read FPower write FPower;
    property Time:TDSTime read FTime write FTime;
    property Parent:TDSDevice read FParent write FParent;
    property Param:Pointer read FParam write FParam;
    destructor Destroy; override;
  end;

  TTransactRecord=record
    Transact: TTransact;
    Time: TDSTime;
  end;

  TTransactList=class(TList)
  private
    function GetItems(Index: integer): TTransact;
    procedure SetItems(Index: integer; const Value: TTransact);
    function GetTime(Index: integer): TDSTime;
    procedure SetTime(Index: integer; const Value: TDSTime);
  public
    function Add(Transact: TTransact; Time: TDSTime = 0): Integer;
    procedure Clear;override;
    procedure Reset;dynamic;
    function First: TTransact;
    function IndexOf(Transact: TTransact): Integer;
    procedure Insert(Index: Integer; Transact: TTransact; Time: TDSTime = 0);
    function Last: TTransact;
    function Remove(Transact: TTransact): Integer;
    function RemoveTime(Transact: TTransact): TDSTime;
    destructor Destroy; override;           // 
    function TotalPower: Integer;
    property Items[Index:integer]:TTransact read GetItems write SetItems; default;
    property Times[Index:integer]:TDSTime read GetTime write SetTime; //,   
  end;

  TQueue=class(TTransactList)
  private
    FMaxCount: integer;           // 
    FSumLengthTime: double;    //     
    FSumTime: double;          //    
    FTransactCount: integer;      //  
    FZeroCount: integer;          //  
    FLastChangeTime: TDSTime;  //   (  )
    FOwner: TDSDevice;         // -  
    function GetAvgCount: double;
    function GetAvgTime: double;
  protected
    procedure Change; dynamic;
  public
    constructor Create(AOwner: TDSDevice);
    procedure Clear;override;
    procedure Add(Transact: TTransact);       // -  
    function Extract: TTransact;               //  
    property MaxLength:integer read FMaxCount;    //.  
    property ZeroCount:integer read FZeroCount;   //  
    property AverageLength:double read GetAvgCount;//  
    property AverageTime:double read GetAvgTime;   //   -  
  end;

  TDeviceList=class(TList)
  private
    function GetItems(Index: integer): TDSDevice;
    procedure SetItems(Index: integer; const Value: TDSDevice);
  public
    function Add(ADevice: TDSDevice): Integer;
    function First: TDSDevice;
    function IndexOf(ADevice: TDSDevice): Integer;
    procedure Insert(Index: Integer; ADevice: TDSDevice);
    function Last: TDSDevice;
    function Remove(ADevice: TDSDevice): Integer;
    property Items[Index:integer]:TDSDevice read GetItems write SetItems; default;
  end;

  TTimeEventMethod = procedure (Sender: TObject; Transact:TTransact; Device: TDSDevice; Id: integer) of object;
  TTimeEvent = record
      Time:TDSTime;         //,    
      Transact:TTransact;   //,     ( Nil)
      Device:TDSDevice;     //,     ( Nil)
      Method:TTimeEventMethod; //,     ( Nil)
      Id: integer; 		//  (  0)
  end;
  PTimeEvent=^TTimeEvent;

  TTimeEventList=class(TList) // ,     
  private
    FOwner: TDSTimer;
    function GetItems(Index: integer): TTimeEvent;
    procedure SetItems(Index: integer; const Value: TTimeEvent);
    function Add(ATimeEvent: TTimeEvent): Integer;overload;
  public
    constructor Create(AOwner: TDSTimer);
    function Add(Interval:TDSTime;Transact:TTransact=nil;Device:TDSDevice=nil;Id: integer=0; Method:TTimeEventMethod=nil):integer;overload;
    function Cancel(Device:TDSDevice):TTimeEvent;//   ( )   
    function First: TTimeEvent;
    procedure Insert(Index: Integer; ATimeEvent: TTimeEvent);
    function Last: TTimeEvent;
    procedure Show(Strings: TStrings);
    destructor Destroy; override;
    property Items[Index:integer]: TTimeEvent read GetItems write SetItems; default;
  end;

  TTransactEvent = procedure (Sender: TObject;Transact:TTransact) of object;

  TDSTimerThread = class(TThread)
  private
    FOwner: TDSTimer;
    FSleepTime: integer;  // ( )
  protected
    procedure Execute; override;
  public
    constructor Create(ACreateSuspended: Boolean; AOwner: TDSTimer);
  end;

  TDSTimer=class(TComponent)
  private
    FEventList:TTimeEventList;       // 
    FDeviceList:TDeviceList;         // ,   
    FFinishTime: TDSTime;               //  
    FResetTime: TDSTime;             //  
    FCurTime: TDSTime;               //  
    FTimerThread: TDSTimerThread;
    FOnStop: TNotifyEvent;
    FOnReset: TNotifyEvent;
    FOnTimer: TNotifyEvent;
    FOnTimeEvent: TTimeEventMethod;
    FTimeScale: integer;
    FWaitForExit: boolean;                // - 1  : 1 .
    FFinished: boolean; 
    procedure SetFinishTime(const Value: TDSTime);
    function GetCurTime: TDSTime;
    procedure Reset2(Sender: TObject; Transact: TTransact;
      Device: TDSDevice;Id:integer);
    procedure Finish(Sender: TObject; Transact: TTransact;
      Device: TDSDevice;Id:integer);
  protected
    procedure DoStop;dynamic;
    procedure ProcessEvents;dynamic;
  public
    procedure Start;          // 
    procedure Stop;           // 
    procedure Pause;          //  
    procedure Reset;          //    ,     
    procedure Clear;          //    ,    
    constructor Create(AOwner:TComponent); override;
    procedure Report(Strings:TStrings);dynamic;//  
    destructor Destroy;override;
    procedure AddDevice(const Value: TDSDevice);
    procedure RemoveDevice(const Value: TDSDevice);
    property CurTime:TDSTime read GetCurTime;
    property Events: TTimeEventList read FEventList;
  published
    property FinishTime:TDSTime read FFinishTime write SetFinishTime default 1000;
    property ResetTime:TDSTime read FResetTime write FResetTime default 0;
    property TimeScale:integer read FTimeScale write FTimeScale default 10;
    property OnReset: TNotifyEvent read FOnReset write FOnReset;
    property OnFinish: TNotifyEvent read FOnStop write FOnStop;
    property OnTimer: TNotifyEvent read FOnTimer write FOnTimer;
    property OnTimeEvent: TTimeEventMethod read FOnTimeEvent write FOnTimeEvent;
    property WaitForExit: boolean read FWaitForExit write FWaitForExit;
  end;

  TGetTimeEvent = procedure (Sender: TObject; Transact:TTransact; var Interval:TDSTime) of object;
  TDSEnterDevice=class;

  TDSDevice=class(TGraphicControl)
  private
    FOnGetInterval: TGetTimeEvent;//     
    FTimer: TDSTimer;            //
    FNext: TDSEnterDevice;          //    ()
    FAlternative: TDSEnterDevice;          //    ()
    FAmount: integer;              // () -
    FBeforeLeave: TTransactEvent;
    FAfterLeave: TNotifyEvent;
    FIntervalMax: TDSTime;       //     
    FIntervalMin: TDSTime;
    FOnRefuse: TTransactEvent;       // 
    procedure SetNext(const Value: TDSEnterDevice);
    procedure SetTimer(const Value: TDSTimer);
    procedure SetIntervalMax(const Value: TDSTime);
    procedure SetIntervalMin(const Value: TDSTime);
    procedure CMTextChanged(var Message: TMessage);message CM_TEXTCHANGED;// 
  protected
    function DoBeforeLeave(Transact: TTransact): boolean;dynamic;
    procedure DoAfterLeave;dynamic;
    procedure DoLeave(Transact: TTransact);dynamic;
    function DoRefuse(Transact: TTransact): boolean;dynamic;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure Paint;override;
    function GetInterval(Transact:TTransact):TDSTime;virtual;// ,  . 
    property OnIntervalQuery:TGetTimeEvent read FOnGetInterval write FOnGetInterval;
    property IntervalMin:TDSTime read FIntervalMin write SetIntervalMin default 10;
    property IntervalMax:TDSTime read FIntervalMax write SetIntervalMax default 10;
    property Destination:TDSEnterDevice read FNext write SetNext;
    property Alternative:TDSEnterDevice read FAlternative write FAlternative;
    property BeforeLeave:TTransactEvent read FBeforeLeave write FBeforeLeave;
    property AfterLeave:TNotifyEvent read FAfterLeave write FAfterLeave;
    property OnRefuse:TTransactEvent read FOnRefuse write FOnRefuse;
    property Timer: TDSTimer read FTimer write SetTimer;
  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    procedure Clear;virtual;                    // ,   . 
    procedure Reset;virtual;                    // 
    procedure Leave(Transact:TTransact);virtual;
    procedure Repaint;override;
    property Amount:integer read FAmount;
  published
    property Caption;
    property Color;
    property Enabled;
    property ShowHint;
    property Visible;
    property OnClick;
    property OnDblClick;
{    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
}
    property Width default 100;
    property Height default 40;
  end;

  TDSGenerator=class(TDSDevice)
  private
    FLimit: integer;          //  -
    FPriority: integer;       //  -
    FPower: integer;          //    -
    FOffset: TDSTime;      //    
  protected
    procedure Paint;override;
  public
    constructor Create(AOwner:TComponent); override;
    procedure Leave(Transact: TTransact);override;
  published
    property Destination;
    property Limit: integer read FLimit write FLimit default 0;
    property FirstOffset: TDSTime read FOffset write FOffset  default 0;
    property Power:integer read FPower write FPower default 1;
    property Priority:integer read FPriority write FPriority default 0;
    property OnIntervalQuery;
    property IntervalMin;
    property IntervalMax;
    property Timer;
    property BeforeLeave;
    property AfterLeave;
    property OnRefuse;
    property Alternative;
  end;

  TDSEnterDevice=class(TDSDevice)           //,  
  private
    FBeforeEnter: TTransactEvent;
    FClosed: boolean;
    procedure SetClosed(const Value: boolean);
  protected
    function DoBeforeEnter(Transact: TTransact):boolean;dynamic;
    property BeforeEnter:TTransactEvent read FBeforeEnter write FBeforeEnter;
    procedure Paint;override;
  public
    function Enter(Transact:TTransact): boolean;virtual;
    procedure Close;virtual; //   
    procedure Open;virtual;  // 
  published
    property Closed:boolean read FClosed write SetClosed default False;
  end;

  TDSAdvance=class(TDSEnterDevice)           //,  
  private
    FCurCount: integer;
  protected
    procedure Paint;override;
  public
    constructor Create(AOwner:TComponent); override;
    function Enter(Transact:TTransact): boolean;override;
    procedure Leave(Transact: TTransact);override;
    procedure Reset;override;
  published
    property CurCount: integer read FCurCount;
    property IntervalMin;
    property IntervalMax;
    property Destination;
    property Timer;
    property BeforeEnter;
    property BeforeLeave;
    property AfterLeave;
    property OnIntervalQuery;
    property OnRefuse;
    property Alternative;
  end;

  TDSTerminator=class(TDSEnterDevice)
  private
    FLimit: integer;
  protected
    procedure Paint;override;
  public
    constructor Create(AOwner: TComponent);override;
    function Enter(Transact:TTransact): boolean;override;
  published
    property BeforeEnter;
    property Limit: integer read FLimit write FLimit default 0;
    property Timer;
    property Height default 40;
  end;

  TCancelRecord=record
    Transact:TTransact;
    Time:TDSTime;
  end;

  TDSChannelDevice=class(TDSEnterDevice)         // 
  private
    FBeforeReserve: TTransactEvent;
    FBeforeEnqueue: TTransactEvent;
    FBeforeExtract: TTransactEvent;
    FPower: integer;                //  
    FSumTime: double;            //  
    FServisedTransacts: TTransactList; // 
    FReleasingDevice: TDSChannelDevice;
    FReleasedByMe: TDeviceList;  //    
    function GetAverageUtil: double; //   %
    function GetAverageTime: double; //  
    function GetFree: integer;          //  
    procedure RemoveReleasedDevice(Device: TDSChannelDevice);
    procedure SetReleasingDevice(const Value: TDSChannelDevice);
    procedure ClearReleasingDevice;
    procedure SetPower(const Value: integer);
  protected
    function DoBeforeReserve(Transact: TTransact): boolean;dynamic;
    function DoBeforeEnqueue(Transact: TTransact): boolean;dynamic;
    function DoBeforeExtract(Transact: TTransact): boolean;dynamic;
    procedure Enqueue(Transact: TTransact);virtual;abstract; //   -  
    procedure Extract;virtual;abstract; //   - ?
    procedure Paint;override;
    property FreeChannels:integer read GetFree;
    property Channels:integer read FPower write SetPower default 1;
    property BeforeReserve:TTransactEvent read FBeforeReserve write FBeforeReserve;
    property BeforeEnqueue:TTransactEvent read FBeforeEnqueue write FBeforeEnqueue;
    property BeforeExtract:TTransactEvent read FBeforeExtract write FBeforeExtract;
    property ReleasingDevice:TDSChannelDevice read FReleasingDevice write SetReleasingDevice;
    property ServisedTransacts: TTransactList read FServisedTransacts write FServisedTransacts;
  public
    procedure AddReleasedDevice(Device:TDSChannelDevice);virtual;
    constructor Create(AOwner:TComponent); override;
    function Enter(Transact:TTransact): boolean;override;
    function Cancel:TCancelRecord;virtual;   //  (. TTimeEventList.Cancel)
    function CanReserve(Transact:TTransact):boolean;virtual;
    procedure Clear;override;
    procedure Leave(Transact:TTransact);override;
    procedure Reset;override;
    procedure Reserve(Transact: TTransact);virtual;
    procedure Release(Transact: TTransact);virtual;
    property AverageUtil: double read GetAverageUtil; // 
    property AverageTime: double read GetAverageTime; //  . 1 
    property SumTime: double read FSumTime;           // .  
    destructor Destroy; override;
  published
    property Height default 66;
    property Timer;
    property OnIntervalQuery;
    property IntervalMin;
    property IntervalMax;
    property Destination;
  end;

  TDSQueueChannelDevice=class(TDSChannelDevice)
  private
    FQueue: TQueue;                            //   -
  protected
    procedure Paint;override;
  public
    constructor Create(AOwner:TComponent); override;
    procedure Clear;override;
    destructor Destroy; override;
    procedure Enqueue(Transact: TTransact);override;  //   
    procedure Extract;override;                       //  
    property Queue:TQueue read FQueue;
  published
    property ReleasingDevice;
  end;

  TDSQueue=class(TDSEnterDevice)      // 
  private
    FBeforeEnqueue: TTransactEvent;
    FBeforeExtract: TTransactEvent;
    FDevices: TDeviceList;         // - 
    FQueue: TQueue;
    FLimit: integer;
  protected
    function DoBeforeEnqueue(Transact: TTransact): boolean;dynamic;
    function DoBeforeExtract(Transact: TTransact): boolean;dynamic;
    procedure Paint;override;
  public
    procedure AddDevice(const Device: TDSChannelDevice);
    procedure RemoveDevice(const Device: TDSEnterDevice);
    constructor Create(AOwner: TComponent);override;
    destructor Destroy;override;
    procedure Clear;override;
    procedure Reset;override;
    procedure Enqueue(Transact: TTransact);virtual;
    function Enter(Transact:TTransact): boolean;override;
    function Extract: TTransact;virtual;
    property Queue:TQueue read FQueue;
  published
    property Limit: integer read FLimit write FLimit default 0;
    property Timer;
    property BeforeEnqueue:TTransactEvent read FBeforeEnqueue write FBeforeEnqueue;
    property BeforeExtract:TTransactEvent read FBeforeExtract write FBeforeExtract;
    property BeforeEnter;
    property AfterLeave;
    property Alternative;
    property OnLimit: TTransactEvent read FOnRefuse write FOnRefuse;//!!!!!!!!!
  end;

  TDSQueuelessChannelDevice=class(TDSChannelDevice)
  private
    FQueue: TDSQueue;  //   -
    procedure SetQueue(const Value: TDSQueue);
  protected
    procedure Enqueue(Transact: TTransact);override;
    procedure Extract;override;
    procedure Notification(AComponent: TComponent; Operation: TOperation);override;
    procedure Paint;override;
  public
    function Enter(Transact:TTransact): boolean;override;
  published
    property Queue:TDSQueue read FQueue write SetQueue;
    property ReleasingDevice;
  end;

  TDSQueueOneChannelDevice=class(TDSQueueChannelDevice)
  private
    function GetTransact: TTransact;
  protected
    procedure Paint;override;
  public
    property ServisedTransact: TTransact read GetTransact;
  published
    property BeforeEnqueue;
    property BeforeEnter;
    property BeforeExtract;
    property BeforeReserve;
    property BeforeLeave;
    property AfterLeave;
    property OnRefuse;
  end;

  TDSQueueMultiChannelDevice=class(TDSQueueChannelDevice)
  protected
    procedure Paint;override;
  public
    property FreeChannels;
    property ServisedTransacts;
    constructor Create(AOwner:TComponent); override;
  published
    property Channels default 3;
    property BeforeEnter;
    property BeforeExtract;
    property BeforeReserve;
    property BeforeEnqueue;
    property BeforeLeave;
    property AfterLeave;
    property OnRefuse;
  end;

  TDSOneChannelDevice=class(TDSQueuelessChannelDevice)
  private
    function GetTransact: TTransact;
  protected
    procedure Paint;override;
  public
    property ServisedTransact: TTransact read GetTransact;
  published
    property BeforeEnter;
    property BeforeReserve;
    property BeforeLeave;
    property AfterLeave;
    property OnRefuse;
  end;

  TDSMultiChannelDevice=class(TDSQueuelessChannelDevice)
  protected
    procedure Paint;override;
  public
    property FreeChannels;
    property ServisedTransacts;
    constructor Create(AOwner:TComponent); override;
  published
    property Channels default 3;
    property BeforeEnter;
    property BeforeReserve;
    property BeforeLeave;
    property AfterLeave;
    property OnRefuse;
  end;

  TDestinationQueryEvent = procedure (Sender: TObject;var Destination: TDSEnterDevice; Transact: TTransact) of object;
  TDSTransfer=class(TDSEnterDevice)
  private
    FProbability: real;
    Counts: array [0..1]of integer;
    FOnDestinationQuery: TDestinationQueryEvent;
    procedure SetProbability(const Value: real);
    function GetCount(Index: integer): integer;
  protected
    procedure Paint;override;
  public
    constructor Create(AOwner:TComponent); override;
    function Enter(Transact:TTransact): boolean;override;
    procedure Reset;override;
  published
    property AlternativeCount: integer index 1 read GetCount;
    property DestinationCount: integer index 0 read GetCount;
    property Alternative;
    property Destination;
    property Probability:real read FProbability write SetProbability;
    property OnDestinationQuery:TDestinationQueryEvent read FOnDestinationQuery write FOnDestinationQuery;
    property BeforeLeave;
    property BeforeEnter;
    property Timer;
    property Height default 66;
  end;

  TSplitEvent=procedure (Sender: TObject;Transact: TTransact; Index: integer) of object;
  TDSSplitter=class(TDSEnterDevice)
  private
    FSplitNumber: integer;
    FOnSplit: TSplitEvent;
    procedure SetSplitNumber(const Value: integer);
  protected
    procedure Paint;override;
  public
    constructor Create(AOwner:TComponent); override;
    function Enter(Transact:TTransact): boolean;override;
  published
    property BeforeLeave;
    property BeforeEnter;
    property Destination;
    property Number:integer read FSplitNumber write SetSplitNumber default 3;
    property OnSplit: TSplitEvent read FOnSplit write FOnSplit;
    property Timer;
    property Height default 50;
    property OnRefuse;
    property Alternative;
  end;

  TAssembleEvent=procedure (Sender: TObject; Transact1,Transact2: TTransact; var CanAssemble: boolean) of object;
  TDSAssembler=class(TDSEnterDevice)
  private
    FAssembleNumber: integer;
    FAssemblingTransacts: TTransactList;
    FOnAssembleQuery: TAssembleEvent;
    procedure SetAssembleNumber(const Value: integer);
  protected
    procedure Paint;override;
  public
    procedure Clear;override;
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    function Enter(Transact:TTransact): boolean;override;
  published
    property Number:integer read FAssembleNumber write SetAssembleNumber default 3;
    property BeforeLeave;
    property BeforeEnter;
    property OnAssembleQuery: TAssembleEvent read FOnAssembleQuery write FOnAssembleQuery;
    property Timer;
    property Destination;
    property Height default 50;
    property OnRefuse;
    property Alternative;
  end;

function IRandom(Min,Max:TDSTime):TDSTime; //   Min  Max
function NegExp(AverageTime:TDSTime):TDSTime; // 
function EmpDistribution(Probabls: array of real):integer; // 

procedure Register;

implementation

function TimeEvent(ATime:TDSTime;ADevice:TDSDevice;Transact:TTransact;AMethod:TTimeEventMethod;Id:integer):TTimeEvent;forward;
procedure DrawBusyBar(const Canvas: TCanvas; const DrawRect: TRect; AMax,ACur:integer);forward;
procedure DrawRomb(const Canvas: TCanvas; const DrawRect: TRect);forward;
function CreateError(const ErrCode: integer; const Device: TDSDevice = nil):EDSError;forward;

{=============== TTransact ===============}

constructor TTransact.Create;
begin
   FPower:=1;
end;

destructor TTransact.Destroy;
begin
   if FParam<>nil then
     dispose(FParam);
   FParam:=nil;
end;

procedure TTransact.Assign(Source: TTransact);
begin
  if Source <> nil then
  begin
    Self.FTag:=Source.Tag;
    Self.FPriority:=Source.Priority;
    Self.FPower:=Source.Power;
    Self.FTime:=Source.Time;
    Self.FParam:=Source.Param;//!!!
    Self.FParent:=Source.Parent;
  end;
end;

{===============  TTransactList =============== }

function TTransactList.Add(Transact: TTransact; Time: TDSTime = 0): Integer;
var P:^TTransactRecord;
begin
  new(P);
  P.Transact:=Transact;
  P.Time:=Time;
  Result:=inherited Add(P);
end;

procedure TTransactList.Clear;
var i:integer;
begin
  for i:=0 to Count-1 do
    dispose(inherited Items[i]);
{  // 
  for i:=0 to Count-1 do
  begin
    try
      try
        Items[i].Free;
      finally
        dispose(inherited Items[i]);
      end;
    except
    end;
  end;
}
  inherited Clear;
end;

procedure TTransactList.Reset;
var i:integer;
begin
  for i:=0 to Count-1 do
    Times[i]:=0;
end;

destructor TTransactList.Destroy;
begin
  // 
  Clear;
  inherited Destroy;
end;

function TTransactList.First: TTransact;
begin
  Result:=GetItems(0);
end;

function TTransactList.GetItems(Index: integer): TTransact;
begin
  Result:=TTransactRecord(inherited Items[Index]^).Transact;
end;

function TTransactList.GetTime(Index: integer): TDSTime;
begin
  Result:=TTransactRecord(inherited Items[Index]^).Time;
end;

function TTransactList.IndexOf(Transact: TTransact): Integer;
begin
  Result := 0;
  while (Result < Count) and (Items[Result] <> Transact)do
    inc(Result);
  if Result = Count then Result := -1;
end;

procedure TTransactList.Insert(Index: Integer; Transact: TTransact; Time: TDSTime = 0);
var P:^TTransactRecord;
begin
  new(P);
  P.Transact:=Transact;
  P.Time:=Time;
  inherited Insert(Index,P);
end;

function TTransactList.Last: TTransact;
begin
  Result:=GetItems(Count-1);
end;

function TTransactList.Remove(Transact: TTransact): Integer;
begin
  Result := IndexOf(Transact);
  if Result <> -1 then Delete(Result);
end;

function TTransactList.RemoveTime(Transact: TTransact): TDSTime;
begin
  if IndexOf(Transact)>=0 then
  begin
    Result:=Times[IndexOf(Transact)];
    Remove(Transact);
  end
  else
    Result:=0;
end;

procedure TTransactList.SetItems(Index: integer; const Value: TTransact);
begin
  TTransactRecord(inherited Items[Index]^).Transact:=Value;
end;

procedure TTransactList.SetTime(Index: integer; const Value: TDSTime);
begin
  TTransactRecord(inherited Items[Index]^).Time:=Value;
end;

function TTransactList.TotalPower: Integer;
var i:integer;
begin
  Result:=0;
  for i := 0 to Count-1 do
  begin
    inc(Result,Items[i].Power);
  end;
end;

{===============  TQueue =============== }

constructor TQueue.Create(AOwner: TDSDevice);
begin
  inherited Create;
  FOwner:=AOwner;
end;

procedure TQueue.Add(Transact: TTransact);
var i:integer;
begin
  if (FOwner.FTimer=nil)then
    raise CreateError(7,FOwner);

  Change;
  i:=Count-1;
  //   -   
  while (i>=0)and (Items[i].FPriority < Transact.Priority)do
    dec(i);
  Insert(i+1,Transact,FOwner.FTimer.CurTime);
  // .  (   !!!)
  if Count > FMaxCount then
     FMaxCount:=Count;
end;

function TQueue.Extract: TTransact;
begin
  try
    Result := GetItems(0);
    with FOwner do
      if Timer<>nil then
      begin
        FSumTime:=FSumTime+(Timer.CurTime-Times[0]);
        if (Timer.CurTime-Times[0])=0 then
        begin
          inc(FZeroCount);
          // .  (   )
          if Count = FMaxCount then
             dec(FMaxCount);
        end;
        Change;
        Delete(0);
        inc(FTransactCount);
      end;
  except
//    Result := nil;
    raise CreateError(1,Self.FOwner);
  end
end;

procedure TQueue.Change;
begin
  FSumLengthTime:=FSumLengthTime+Count*(FOwner.Timer.CurTime-FLastChangeTime);
  FLastChangeTime:=FOwner.Timer.CurTime;
end;

procedure TQueue.Clear;
begin
  FMaxCount:=0;
  FTransactCount:=0;
  FSumLengthTime:=0;
  FSumTime:=0;
  FZeroCount:=0;
  FLastChangeTime:=0;
  inherited Clear;
end;

function TQueue.GetAvgTime: double;
begin
  if (FTransactCount-FZeroCount)>0 then
    Result:=FSumTime/(FTransactCount-FZeroCount)
  else
    Result:=0;
end;

function TQueue.GetAvgCount: double;
begin
  Change;
  if FOwner.Timer.CurTime>0 then
    Result:=FSumLengthTime/FOwner.Timer.CurTime
  else
    Result:=0;
end;

{===============  TDeviceList =============== }

function TDeviceList.Add(ADevice: TDSDevice): Integer;
begin
  Result:=inherited Add(ADevice);
end;

function TDeviceList.First: TDSDevice;
begin
  Result:=GetItems(0);
end;

function TDeviceList.GetItems(Index: integer): TDSDevice;
begin
  Result:=inherited Items[Index];
end;

function TDeviceList.IndexOf(ADevice: TDSDevice): Integer;
begin
  Result:=inherited IndexOf(ADevice);
end;

procedure TDeviceList.Insert(Index: Integer; ADevice: TDSDevice);
begin
  inherited Insert(Index,ADevice);
end;

function TDeviceList.Last: TDSDevice;
begin
  Result:=GetItems(Count-1);
end;

function TDeviceList.Remove(ADevice: TDSDevice): Integer;
begin
  Result:=inherited Remove(ADevice);
end;

procedure TDeviceList.SetItems(Index: integer; const Value: TDSDevice);
begin
  inherited Items[Index]:=Value;
end;

{===============  TTimeEventList =============== }

function TTimeEventList.Add(ATimeEvent: TTimeEvent): Integer;
var i:integer;
    Prior: integer;
begin
  if ATimeEvent.Transact <> nil then
    Prior:=ATimeEvent.Transact.Priority
  else
    Prior:=0;
  i:=Count-1;
  //     
  while (i>=0)and((Items[i].Time > ATimeEvent.Time)or(Items[i].Time = ATimeEvent.Time)and(Items[i].Transact<>nil)and(Items[i].Transact.Priority < Prior))do
    dec(i);
  inc(i);
  Insert(i,ATimeEvent);
  Result:=i;
end;

function TTimeEventList.Add(Interval: TDSTime; Transact: TTransact=nil;
  Device: TDSDevice=nil;Id: integer=0;Method:TTimeEventMethod=nil): integer;
begin
  if FOwner<>nil then
  begin
    Result:=Add(TimeEvent(FOwner.FCurTime+Interval,Device,Transact,Method,Id));
  end
  else
  begin
//    Result := -1;
    raise CreateError(4);
  end;
end;

function TTimeEventList.Cancel(Device: TDSDevice): TTimeEvent;
var i:integer;
begin
  for i:=Count-1 downto 0 do
    if Items[i].Device=Device then
    begin
      Result:=Items[i];
      Result.Time:=Result.Time-FOwner.CurTime;
      Delete(i);
      Exit;
    end;
  raise CreateError(13,Device);
end;

constructor TTimeEventList.Create(AOwner: TDSTimer);
begin
  FOwner:=AOwner;
end;

destructor TTimeEventList.Destroy;
var i:integer;
begin
  for i:=0 to Count-1 do
  begin
    Items[i].Transact.Free;
    dispose(inherited Items[i]);
  end;
  inherited Destroy;
end;

function TTimeEventList.First: TTimeEvent;
begin
  Result:=GetItems(0);
end;

function TTimeEventList.GetItems(Index: integer): TTimeEvent;
begin
  Result:=TTimeEvent(inherited Items[Index]^);
end;

procedure TTimeEventList.Insert(Index: Integer; ATimeEvent: TTimeEvent);
var P:PTimeEvent;
begin
  new(P);
  P^:=ATimeEvent;
  inherited Insert(Index,P);
end;

function TTimeEventList.Last: TTimeEvent;
begin
  Result:=GetItems(Count-1);
end;

procedure TTimeEventList.SetItems(Index: integer; const Value: TTimeEvent);
begin
  PTimeEvent(inherited Items[Index])^:=Value;
end;

procedure TTimeEventList.Show(Strings:TStrings);
var i:integer;
    Cap:string;
begin
  with Strings do
  begin
    if FOwner<>nil then
      Add('--- '+IntToStr(FOwner.FCurTime)+' ---');
    for i:=0 to Self.Count-1 do
    begin
      if Items[i].Device<>nil then
        Cap:=Items[i].Device.Name
      else
        Cap:='----';
//      Add(Format('%.5d %.11s %.4d     %.4d',[Items[i].Time,
      Add(Format('%.5d %.2d %p %.11s',[Items[i].Time,Items[i].Id,Pointer(Items[i].Transact),Cap]));
//        Items[i].Transact.Priority,
//        Items[i].Transact.Tag]));
    end;
  end;
end;

{===============  TDSTimerThread ===============}

constructor TDSTimerThread.Create(ACreateSuspended: Boolean;
  AOwner: TDSTimer);
begin
  inherited Create(ACreateSuspended);
  FOwner:=AOwner;
  FreeOnTerminate:=True;
end;

procedure TDSTimerThread.Execute;
begin
  while(not Terminated)do
  begin
    Synchronize(FOwner.ProcessEvents);
    Sleep(FSleepTime);
  end;
end;

{===============  TDSTimer ===============}

constructor TDSTimer.Create(AOwner:TComponent);
begin
   inherited Create(AOwner);
   FEventList:=TTimeEventList.Create(Self);
   FDeviceList:=TDeviceList.Create;
   FTimeScale:=10;
   FFinishTime:=1000;
end;

destructor TDSTimer.Destroy;
begin
  FEventList.Free;
  FDeviceList.Free;
  inherited Destroy;
end;

procedure TDSTimer.Start;
var i:integer;
    Transact: TTransact;
begin
  if FTimerThread<>nil then
    Stop;

  FCurTime:=0;
  FFinished:=False;

  //  
//  if FCurTime=0 then      // 
    for i:=0 to FDeviceList.Count-1 do
       if (FDeviceList[i]is TDSGenerator) then
          FDeviceList[i].Leave(nil);

  Transact:= TTransact.Create;
  Transact.FPriority:=Low(Transact.FPriority);
  if (FResetTime<>0) then
    FEventList.Add(FResetTime,Transact,nil,0,Reset2);

  Transact:= TTransact.Create;
  Transact.FPriority:=Low(Transact.FPriority);
  if (FFinishTime<>0) then
    FEventList.Add(FFinishTime,Transact,nil,0,Finish);

  FTimerThread:=TDSTimerThread.Create(False,Self);
end;

procedure TDSTimer.Stop;
begin
  FTimerThread.Terminate;
  FTimerThread:=nil;
  DoStop;
end;

procedure TDSTimer.DoStop;
var i:integer;
begin
  if TimeScale=0 then
    for i:=0 to FDeviceList.Count-1 do
       FDeviceList[i].Invalidate;
end;

procedure TDSTimer.Finish(Sender: TObject; Transact: TTransact; Device: TDSDevice; Id: integer);
begin
  FFinished:=True;
  Transact.Free;
  if not FWaitForExit then
    Stop;
  if Assigned(FOnStop) then
     FOnStop(Self);
end;

procedure TDSTimer.ProcessEvents;
begin
  if(FEventList.Count=0)then
  begin
    Stop;
    Exit;
  end;

  FCurTime:=FEventList.First.Time;

  if Assigned(FOnTimer)then
    FOnTimer(Self);

  //      
  with FEventList do
  begin
    while (Count>0)and(First.Time=FCurTime)do
    begin
      if Assigned(FOnTimeEvent)then
        FOnTimeEvent(Self,First.Transact,First.Device,First.Id);

      if(not FFinished or not(First.Device is TDSGenerator))then
      begin
        if Assigned(First.Method)then      //  ,   
          First.Method(Self,First.Transact,First.Device,First.Id)
        else
          if(First.Id=0)and(Assigned(First.Device))then     //   
            First.Device.Leave(First.Transact);
      end;
      Delete(0);
    end;
    if(Count>0)and(FTimerThread<>nil)then //    
      FTimerThread.FSleepTime:=(First.Time-FCurTime)*TimeScale;
  end;
end;

procedure TDSTimer.Report(Strings: TStrings);
var i:integer;
begin
  Strings.Add('  = '+IntToStr(FCurTime));
  with FDeviceList do
    for i:=0 to Count-1 do
    begin
      Strings.Add('==== '+ Items[i].Caption+' ====');
      if Items[i] is TDSChannelDevice then
        with TDSQueueChannelDevice(Items[i])do
          Strings.Add(Format(': %u'#9'.: %.2f'#9'.: %.2f%%',[Amount,GetAverageTime,GetAverageUtil]));

      if Items[i] is TDSQueueChannelDevice then
        with TDSQueueChannelDevice(Items[i])do
          Strings.Add(Format(' => .: %.2f'#9'.: %.2f'#9' : %d',[Queue.AverageTime,Queue.AverageLength,Queue.ZeroCount]));

      if Items[i] is TDSQueue then
        with TDSQueue(Items[i])do
          Strings.Add(Format('.: %.2f'#9'.: %.2f'#9' : %d',[Queue.AverageTime,Queue.AverageLength,Queue.ZeroCount]));
    end;
end;

procedure TDSTimer.Pause;
begin
  FTimerThread.Suspended:=not FTimerThread.Suspended;
end;

procedure TDSTimer.Reset;
var i:integer;
begin
  for i:=0 to FDeviceList.Count-1 do
    TDSDevice(FDeviceList[i]).Reset;
  if Assigned(FOnReset) then
     FOnReset(Self);
end;

procedure TDSTimer.Reset2(Sender: TObject; Transact: TTransact; Device: TDSDevice; Id: integer);
begin
  Transact.Free;
  Reset;
end;

procedure TDSTimer.Clear;
var i:integer;
begin
  for i:=0 to FDeviceList.Count-1 do
     FDeviceList[i].Clear;
  FEventList.Clear;//  
  FCurTime:=0;
end;

procedure TDSTimer.SetFinishTime(const Value: TDSTime);
begin
  if Value > FCurTime then
     FFinishTime := Value
  else
     raise CreateError(2);
end;

function TDSTimer.GetCurTime: TDSTime;
begin
  if (FCurTime<FResetTime)then
    Result:=FCurTime
  else
    Result:=FCurTime-FResetTime;
end;

procedure TDSTimer.AddDevice(const Value: TDSDevice);
begin
  if FDeviceList.IndexOf(Value)<0 then
    FDeviceList.Add(Value);
end;

procedure TDSTimer.RemoveDevice(const Value: TDSDevice);
begin
  FDeviceList.Remove(Value);
end;

{===============  TDSDevice =============== }

constructor TDSDevice.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle:=ControlStyle + [csOpaque];
  Width:=100;
  FIntervalMax:=10;
  FIntervalMin:=10;
end;

destructor TDSDevice.Destroy;
var i:integer;
begin
  if FTimer<>nil then begin
    // 
    FTimer.RemoveDevice(Self);

    // 
    with FTimer.Events do
      for i:=Count-1 downto 0 do
      begin
        if Items[i].Device=Self then
          Delete(i);
      end;
  end;
  inherited Destroy;
end;

function TDSDevice.GetInterval(Transact:TTransact): TDSTime;
begin
  if FIntervalMin>FIntervalMax then
    raise CreateError(3,Self);

  if FIntervalMin=FIntervalMax then
    Result:=FIntervalMin
  else
    Result:=IRandom(FIntervalMin,FIntervalMax);

  if Assigned(FOnGetInterval)then
    FOnGetInterval(Self,Transact,Result);
end;

function TDSDevice.DoBeforeLeave(Transact: TTransact): boolean;
begin
  Result:=True;
  try //    
    if Assigned(FBeforeLeave) then
      FBeforeLeave(Self,Transact);
  except
    Result:=False;
  end;
end;

function TDSDevice.DoRefuse(Transact: TTransact): boolean;
begin
  Result:=True;
  try //    
    if Assigned(FOnRefuse) then
      FOnRefuse(Self,Transact);
    Transact.Free;//!!!!!!
    {$IFNDEF SILENCE_TERM}
    if Transact<>nil then
      raise CreateError(12,Self);
    {$ENDIF}
  except
    Result:=False;
  end;
end;

procedure TDSDevice.DoAfterLeave;
begin
  if Assigned(FAfterLeave) then
    FAfterLeave(Self);
end;

procedure TDSDevice.Leave(Transact: TTransact);
begin
  if DoBeforeLeave(Transact)then //    
    DoLeave(Transact);
  inc(FAmount);
  Repaint;
  DoAfterLeave;
end;

procedure TDSDevice.DoLeave(Transact: TTransact);
begin
  if Assigned(FNext) then
  begin
     if not FNext.Enter(Transact)and(not Assigned(FAlternative)or
        not FAlternative.Enter(Transact))
     then DoRefuse(Transact);
  end
  else
  begin
     Transact.Free;                  // -.  
    {$IFNDEF SILENCE_TERM}
     raise CreateError(5,Self);
    {$ENDIF}
  end;
end;

procedure TDSDevice.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent,Operation);
  if (Operation=opRemove) then begin
     if(AComponent=FTimer) then
       FTimer:=nil;
     if(AComponent=FNext) then
       FNext:=nil;
  end;
end;

procedure TDSDevice.Clear;
begin
   Reset;
end;

procedure TDSDevice.Reset;
begin
  FAmount:=0;
  Invalidate;
end;

procedure TDSDevice.Paint;
begin
{  if (FNext<>nil)and(Parent<>nil)and(Parent is TCustomForm)then
    with TCustomForm(Parent).Canvas do
    begin
      Pen.Color:=clBlack;
//      Pen.Width:=3;
      PenPos:=Point(Left+Width div 2, Top+Height);
      with FNext do
        LineTo(Left+Width div 2, Top)
    end;
    }
  with Canvas,ClientRect do
  begin
    Font:=Self.Font;
    Pen.Color:=clBlack;
    Pen.Width:=2;
    Brush.Color:=Self.Color;
  end;
end;

procedure TDSDevice.Repaint;
begin
  if (FTimer=nil)or(FTimer<>nil)and(FTimer.TimeScale>0) then
    inherited Repaint;
end;

procedure TDSDevice.CMTextChanged(var Message: TMessage);
begin
  Invalidate;
end;

procedure TDSDevice.SetIntervalMax(const Value: TDSTime);
begin
  FIntervalMax := Value;
  Invalidate;
end;

procedure TDSDevice.SetIntervalMin(const Value: TDSTime);
begin
  FIntervalMin := Value;
  Invalidate;
end;


procedure TDSDevice.SetNext(const Value: TDSEnterDevice);
begin
  if FNext <> Value then
  begin
    if FNext=Self then
      raise CreateError(6,Self);
    FNext := Value;
    Invalidate;
  end;
end;

procedure TDSDevice.SetTimer(const Value: TDSTimer);
begin
  if FTimer <> Value then
  begin
    if FTimer<>nil then
      FTimer.RemoveDevice(Self);
    if Value<>nil then
      Value.AddDevice(Self);
    FTimer := Value;
    Invalidate;
  end;
end;

{===============  TDSGenerator =============== }

constructor TDSGenerator.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FPower:=1;
  Height:=40;
end;

procedure TDSGenerator.Paint;
begin
  inherited Paint;
  with ClientRect,Canvas do
  begin
    RoundRect(Left,Top,Right,Bottom,(Bottom-Top),(Bottom-Top));
    SetTextAlign(Handle, ta_Center or ta_Top);
    TextOut((Right-Left)div 2,3,Caption);
    TextOut((Right-Left)div 2,(Bottom-Top)div 2 + 3,IntToStr(FAmount));
  end;
end;

procedure TDSGenerator.Leave(Transact: TTransact);
var Offs:TDSTime;
begin
  if FTimer=nil then
    raise CreateError(7,Self);

  //  ?
  if (Transact<>nil)then
  begin
    Transact.Time:=FTimer.CurTime;
    inherited Leave(Transact);
    Offs:=0;
  end
  else
    Offs:=FOffset;

  //   (FLimit=0 - )
  if ((FLimit>0)and(FAmount=FLimit))or(FTimer=nil)then
    exit;

  //       
  Transact:=TTransact.Create;
  with Transact do
  begin
    Priority:=FPriority;
    Power:=FPower;
    Parent:=Self;
  end;
  FTimer.FEventList.Add(Offs+GetInterval(Transact),Transact,Self);
end;

{=============== TDSEnterDevice ==================}

function TDSEnterDevice.DoBeforeEnter(Transact: TTransact):boolean;
begin
  Result:=True;
  try
    if Assigned(FBeforeEnter) then
      FBeforeEnter(Self,Transact);
  except
    Result:=False;
  end;
end;

function TDSEnterDevice.Enter(Transact: TTransact):boolean;
begin
  Result:=not FClosed and DoBeforeEnter(Transact);
end;

procedure TDSEnterDevice.Close;
begin
  Closed:=True;
end;

procedure TDSEnterDevice.Open;
begin
  Closed:=False;
end;

procedure TDSEnterDevice.SetClosed(const Value: boolean);
begin
  if FClosed <> Value then
  begin
    FClosed := Value;
    Invalidate;
  end;
end;

procedure TDSEnterDevice.Paint;
begin
  inherited Paint;
  if FClosed then
    Canvas.Pen.Color:=clDkGray;
end;

{===============  TDSQueue =============== }

constructor TDSQueue.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FQueue:=TQueue.Create(Self);
  FDevices:=TDeviceList.Create;
  Width:=100;
  Height:=40;
end;

destructor TDSQueue.Destroy;
begin
  Queue.Free;
  FDevices.Free;
  inherited Destroy;
end;

function TDSQueue.Enter(Transact: TTransact):boolean;
var i:integer;
begin
  Result:=inherited Enter(Transact);
  if Result then
  begin
    i:=0;
    while(i<FDevices.Count)and(not TDSChannelDevice(FDevices[i]).Enter(Transact))do
      inc(i);
    if i=FDevices.Count then
      Enqueue(Transact);
  end;
end;

function TDSQueue.DoBeforeEnqueue(Transact: TTransact): boolean;
begin
  Result:=True;
  try //    
    if Assigned(FBeforeEnqueue) then
      FBeforeEnqueue(Self,Transact);
  except
    Result:=False;
  end;
end;

procedure TDSQueue.Enqueue(Transact: TTransact);
begin
  if FTimer=nil then
    raise CreateError(7,Self);
  if DoBeforeEnqueue(Transact)then
  begin
    if (FLimit>0)and(FQueue.Count=FLimit)then // 
    begin
      if not Assigned(FAlternative)or not FAlternative.Enter(Transact)then
        DoRefuse(Transact);
    end
    else
      Queue.Add(Transact);
    Repaint;
  end;
end;

function TDSQueue.DoBeforeExtract(Transact: TTransact): boolean;
begin
  Result:=True;
  try //    
    if Assigned(FBeforeExtract) then
      FBeforeExtract(Self,Transact);
  except
    Result:=False;
  end;
end;

function TDSQueue.Extract:TTransact;
begin
  if DoBeforeExtract(Queue.First)then
  begin
    Result:=Queue.Extract;
    Repaint;
  end
  else
    Result:=nil;
end;

procedure TDSQueue.AddDevice(const Device: TDSChannelDevice);
begin
  FDevices.Add(Device);
end;

procedure TDSQueue.RemoveDevice(const Device: TDSEnterDevice);
begin
  FDevices.Remove(Device);
end;

procedure TDSQueue.Clear;
begin
  Queue.Clear;
  inherited Clear;
end;

procedure TDSQueue.Reset;
begin
  FQueue.Reset;
  inherited Reset;
end;

procedure TDSQueue.Paint;
begin
  inherited Paint;
  with Canvas,ClientRect do
  begin
    Rectangle(Left,Top,Right,Bottom);

    SetTextAlign(Handle, ta_Left or ta_Top);
    TextOut(Left+3,3,Caption);

    DrawBusyBar(Canvas,Rect(Left+5,Bottom-(Bottom-Top)div 3,Right-5,Bottom-5),Queue.MaxLength,Queue.Count);
  end;
end;

{===============  TDSChannelDevice =============== }

constructor TDSChannelDevice.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FReleasedByMe:=TDeviceList.Create;
  FServisedTransacts:=TTransactList.Create;
  Height:=66;
  FPower:=1;
end;

destructor TDSChannelDevice.Destroy;
begin
  ClearReleasingDevice;
  FReleasedByMe.Free;
  FServisedTransacts.Free;
  if FReleasingDevice<>nil then
    FReleasingDevice.RemoveReleasedDevice(Self);
  inherited Destroy;
end;

function TDSChannelDevice.Enter(Transact: TTransact):boolean;
begin
  Result:=inherited Enter(Transact);
  if Result then
    if CanReserve(Transact) then
    begin //     
      Reserve(Transact);
    end
    else
    begin
      Enqueue(Transact);
    end;
end;

procedure TDSChannelDevice.Leave(Transact: TTransact);
var i:integer;
begin
  //   
  for i:=0 to FReleasedByMe.Count-1 do
    if TDSChannelDevice(FReleasedByMe[i]).FServisedTransacts.IndexOf(Transact)>=0 then
    begin
      TDSChannelDevice(FReleasedByMe[i]).Release(Transact);
    end;

  if DoBeforeLeave(Transact)then //    
    DoLeave(Transact);

  if FReleasingDevice=nil then
  begin
    Release(Transact);
  end;
end;

function TDSChannelDevice.DoBeforeReserve(Transact: TTransact): boolean;
begin
  Result:=True;
  try //    
    if Assigned(FBeforeReserve) then
      FBeforeReserve(Self,Transact);
  except
    Result:=False;
  end;
end;

procedure TDSChannelDevice.Reserve(Transact: TTransact);
var i: TDSTime;
begin
  if FPower<Transact.Power then
    raise CreateError(8,Self);

  if (FTimer=nil)then
    raise CreateError(7,Self);

  if DoBeforeReserve(Transact)then //    
  begin
    FServisedTransacts.Add(Transact,FTimer.CurTime);
    i:=GetInterval(Transact);
    if i>=0 then
      FTimer.FEventList.Add(i,Transact,Self);
  end;
  Repaint;
end;

procedure TDSChannelDevice.Release(Transact: TTransact);
var i:integer;
begin
  i:=FServisedTransacts.IndexOf(Transact);
  if i>=0 then
  begin
    FSumTime:=FSumTime+FTimer.CurTime-FServisedTransacts.Times[i];
    FServisedTransacts.Delete(i);
  end
  else
    raise CreateError(9,Self);
  DoAfterLeave;
  inc(FAmount);//!!!!!! Cancel -
  Extract;
  Repaint;
end;

function TDSChannelDevice.DoBeforeEnqueue(Transact: TTransact): boolean;
begin
  Result:=True;
  try //    
    if Assigned(FBeforeEnqueue) then
      FBeforeEnqueue(Self,Transact);
  except
    Result:=False;
  end;
end;

function TDSChannelDevice.DoBeforeExtract(Transact: TTransact): boolean;
begin
  Result:=True;
  try //    
    if Assigned(FBeforeExtract) then
      FBeforeExtract(Self,Transact);
  except
    Result:=False;
  end;
end;

function TDSChannelDevice.Cancel: TCancelRecord;
var TimeEvent:TTimeEvent;
begin
  if (FTimer=nil)then
    raise CreateError(7,Self);

  if (FServisedTransacts.Count>0)then
  begin
    TimeEvent:=FTimer.Events.Cancel(Self);
    Result.Transact:=TimeEvent.Transact;
    Result.Time:=TimeEvent.Time;
    Release(TimeEvent.Transact);
  end;
end;

function TDSChannelDevice.GetFree: integer;
begin
  Result:=FPower-FServisedTransacts.TotalPower;
end;

function TDSChannelDevice.GetAverageUtil: double;
begin
  if (FTimer<>nil)and(FTimer.CurTime>0) then
    Result:=FSumTime/(FTimer.CurTime*FPower)*100
  else
    Result:=0;
end;

function TDSChannelDevice.GetAverageTime: double;
begin
  if FAmount>0 then
    Result:=FSumTime/FAmount
  else
    Result:=0;
end;

procedure TDSChannelDevice.Reset;
begin
  FSumTime:=0;
  FServisedTransacts.Reset;
  inherited Reset;
end;

procedure TDSChannelDevice.ClearReleasingDevice;
var i,j:integer;
begin
  if FReleasingDevice<>nil then
    for i:= 0 to FServisedTransacts.Count-1 do
      FReleasingDevice.FServisedTransacts.Remove(FServisedTransacts[i]);
  // !
  for j:= 0 to FReleasedByMe.Count-1 do
    for i:= 0 to FServisedTransacts.Count-1 do
      TDSChannelDevice(FReleasedByMe[j]).FServisedTransacts.Remove(FServisedTransacts[i]);
end;

procedure TDSChannelDevice.Clear;
begin
  ClearReleasingDevice;
  FServisedTransacts.Clear;
  inherited Clear;
end;

procedure TDSChannelDevice.Paint;
begin
  inherited Paint;
  with Canvas,ClientRect do
  begin
    Rectangle(Left,Top,Right,Bottom);

    SetTextAlign(Handle, ta_Left or ta_Top);
    TextOut(5,3,Caption);

    SetTextAlign(Handle, ta_Right or ta_Top);
    TextOut(Right-5,3,IntToStr(FAmount));
  end;
end;

procedure TDSChannelDevice.AddReleasedDevice(Device: TDSChannelDevice);
begin
  if FReleasedByMe.IndexOf(Device)<0 then
    FReleasedByMe.Add(Device);
end;

procedure TDSChannelDevice.RemoveReleasedDevice(Device: TDSChannelDevice);
begin
  FReleasedByMe.Remove(Device);
end;

procedure TDSChannelDevice.SetReleasingDevice(const Value: TDSChannelDevice);
begin
  if FReleasingDevice<>Value then
  begin
    if FReleasingDevice=Self then
      raise CreateError(10);
    if FReleasingDevice<>nil then
      FReleasingDevice.RemoveReleasedDevice(Self);
    FReleasingDevice := Value;
    if Value <> nil then
      Value.AddReleasedDevice(Self);
  end;
end;

function TDSChannelDevice.CanReserve(Transact: TTransact): boolean;
begin
  //     
  Result:=not FClosed and (GetFree>=Transact.Power);
end;

{===============  TDSQueueChannelDevice =============== }

constructor TDSQueueChannelDevice.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FQueue:=TQueue.Create(Self);
end;

destructor TDSQueueChannelDevice.Destroy;
begin
  FQueue.Free;     //     
  inherited Destroy;
end;

procedure TDSQueueChannelDevice.Enqueue(Transact: TTransact);
begin
  if DoBeforeEnqueue(Transact)then//    
    FQueue.Add(Transact);
  Repaint;
end;

procedure TDSQueueChannelDevice.Extract;
begin
  while (FQueue.Count > 0)and
     (DoBeforeExtract(FQueue.First))and //    
     (GetFree>=FQueue.First.Power)//     
  do
  begin
    Reserve(FQueue.Extract);
  end;
  Repaint;
end;

procedure TDSQueueChannelDevice.Clear;
begin
  FQueue.Clear;
  inherited Clear;
end;

procedure TDSQueueChannelDevice.Paint;
var StepY: integer;
begin
  inherited Paint;
  with Canvas,ClientRect do
  begin
    SetTextAlign(Handle, ta_Left or ta_Bottom);
    StepY:=(Bottom-Top)div 4;
    TextOut(5,StepY*3+3,'Q:');
    DrawBusyBar(Canvas,Rect(Left+5,StepY*3+5,Right-5,Bottom-5),FQueue.MaxLength,FQueue.Count);
  end;
end;

{===============  TDSQueuelessChannelDevice =============== }

procedure TDSQueuelessChannelDevice.Enqueue(Transact: TTransact);
begin
  if (FQueue<>nil)and(DoBeforeEnqueue(Transact)) then
    FQueue.Enqueue(Transact);
end;

procedure TDSQueuelessChannelDevice.Extract;
begin
  while (FQueue<>nil)and(FQueue.Queue.Count > 0)and
     (DoBeforeExtract(FQueue.Queue.First))and
     (CanReserve(FQueue.Queue.First))do
  begin
    Reserve(FQueue.Extract);
  end;
end;

procedure TDSQueuelessChannelDevice.Paint;
begin
  inherited Paint;
  with Canvas,ClientRect do
  begin
    SetTextAlign(Handle, ta_Center or ta_Bottom);
    TextOut((Right-Left)div 2,Bottom-3,Format('%d..%d',[FIntervalMin,FIntervalMax]));
  end;
end;

procedure TDSQueuelessChannelDevice.SetQueue(const Value: TDSQueue);
begin
  if FQueue <> Value then
  begin
    if FQueue <> nil then
      FQueue.RemoveDevice(Self);
    FQueue := Value;
    if FQueue <> nil then
      FQueue.AddDevice(Self);
  end;
end;

procedure TDSQueuelessChannelDevice.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent,Operation);
  if (Operation=opRemove)and(AComponent=FQueue) then
    FQueue:=nil;
end;

{===============  TDSQueueOneChannelDevice =============== }

function TDSQueueOneChannelDevice.GetTransact: TTransact;
begin
  if FServisedTransacts.Count>0 then
    Result:=FServisedTransacts[0]
  else
    Result:=nil;
end;

procedure TDSQueueOneChannelDevice.Paint;
var CenterX, StepY: integer;
begin
  inherited Paint;
  with Canvas,ClientRect do
  begin
    Brush.Color:=Self.Color;
    CenterX:=(Right-Left)div 2;
    StepY:=(Bottom-Top)div 4;

    SetTextAlign(Handle, ta_Left or TA_BOTTOM);
    TextOut(CenterX+StepY, StepY*2,IntToStr(Round(GetAverageUtil))+'%');

    Pen.Width:=1;
    if FServisedTransacts.Count>0 then
       Brush.Color:=clBlue
    else
       Brush.Color:=clNavy;
    Ellipse(CenterX-StepY div 2,StepY,CenterX+StepY div 2,StepY*2);
//    Rectangle(CenterX-StepY div 4,StepY div 2,CenterX+StepY div 4,StepY);
  end;
end;

{===============  TDSQueueMultiChannelDevice =============== }

constructor TDSQueueMultiChannelDevice.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FPower:=3;
end;

procedure TDSQueueMultiChannelDevice.Paint;
var StepY: integer;
begin
  inherited Paint;
  with Canvas,ClientRect do
  begin
    Brush.Color:=Self.Color;
    StepY:=(Bottom-Top)div 4;
    SetTextAlign(Handle, ta_Left or ta_Bottom);
    TextOut(5,StepY*2-4,'B:');
    DrawBusyBar(Canvas,Rect(Left+5,StepY*2-2,Right-5,StepY*3-10),FPower,FServisedTransacts.TotalPower);
  end;
end;

{===============  TDSOneChannelDevice =============== }

function TDSOneChannelDevice.GetTransact: TTransact;
begin
  if FServisedTransacts.Count>0 then
    Result:=FServisedTransacts[0]
  else
    Result:=nil;
end;

procedure TDSOneChannelDevice.Paint;
var CenterX, StepY: integer;
begin
  inherited Paint;
  with Canvas,ClientRect do
  begin
    Brush.Color:=Self.Color;
    CenterX:=(Right-Left)div 2;
    StepY:=(Bottom-Top)div 4;

    SetTextAlign(Handle, ta_Left or TA_BOTTOM);
    TextOut(CenterX+StepY, StepY*2,IntToStr(Round(GetAverageUtil))+'%');

    Pen.Width:=1;
    if FServisedTransacts.Count>0 then
       Brush.Color:=clBlue
    else
       Brush.Color:=clNavy;
    Ellipse(CenterX-StepY div 2,StepY,CenterX+StepY div 2,StepY*2);
//    Rectangle(CenterX-StepY div 4,StepY div 2,CenterX+StepY div 4,StepY);
  end;
end;

{===============  TDSMultiChannelDevice =============== }

constructor TDSMultiChannelDevice.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FPower:=3;
end;

procedure TDSMultiChannelDevice.Paint;
var StepY: integer;
begin
  inherited Paint;
  with Canvas,ClientRect do
  begin
    Brush.Color:=Self.Color;
    StepY:=(Bottom-Top)div 4;
    SetTextAlign(Handle, ta_Left or ta_Bottom);
    TextOut(5,StepY*2-4,'B:');
    DrawBusyBar(Canvas,Rect(Left+5,StepY*2-2,Right-5,StepY*3-10),FPower,FServisedTransacts.TotalPower);
  end;
end;

//=============== TDSTransfer ================

constructor TDSTransfer.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Height:=66;
  Probability:=0.5;
end;

function TDSTransfer.Enter(Transact: TTransact):boolean;
var Device: TDSEnterDevice;
begin
  Result:=inherited Enter(Transact);
  if Result then
  begin
    if Random<Probability then
    begin
      Device:=FNext;
      inc(Counts[0]);
    end
    else
    begin
      Device:=FAlternative;
      inc(Counts[1]);
    end;

    if Assigned(FOnDestinationQuery)then
      FOnDestinationQuery(Self,Device,Transact);

    if Assigned(Device) then
    begin
       if DoBeforeLeave(Transact)then
       begin
         if not Device.Enter(Transact)then
           DoRefuse(Transact);
       end;
    end
    else
    begin
       Transact.Free;                  // -.  
      {$IFNDEF SILENCE_TERM}
       raise CreateError(5,Self);
      {$ENDIF}
    end;
    inc(FAmount);
    DoAfterLeave;
    Repaint;
  end;
end;

procedure TDSTransfer.Reset;
begin
  inherited Reset;
  Counts[0]:=0;
  Counts[1]:=0;
end;

procedure TDSTransfer.Paint;
var CenterX, CenterY: integer;
begin
  inherited Paint;
  with Canvas,ClientRect do
  begin
    CenterX:=(Right-Left)div 2;
    CenterY:=(Bottom-Top)div 2;

    DrawRomb(Canvas,ClientRect);

    SetTextAlign(Handle, ta_Center or ta_Bottom);
    TextOut(CenterX,CenterY-3,Caption);

    SetTextAlign(Handle, ta_Center or ta_Top);
    TextOut(CenterX,CenterY+3,Format('%.2f',[Probability]));

    SetTextAlign(Handle, ta_Left or ta_Bottom);
    TextOut(Left,CenterY-5,IntToStr(Counts[0]));

    SetTextAlign(Handle, ta_Right or ta_Bottom);
    TextOut(Right,CenterY-5,IntToStr(Counts[1]));
  end;
end;

procedure TDSTransfer.SetProbability(const Value: real);
begin
  if (Value>=0)and(Value<=1)then
  begin
    FProbability := Value;
    Invalidate;
  end
  else
    raise CreateError(11,Self);
end;

//=============== TDSTerminator ===============

constructor TDSTerminator.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Height:=40;
end;

function TDSTerminator.Enter(Transact: TTransact):boolean;
begin
  Result:=inherited Enter(Transact);
  if Result then
  begin
    Transact.Free;
    inc(FAmount);
    if (FAmount=FLimit)and(FTimer<>nil)then
       FTimer.Finish(Self,nil,Self,0);
    Repaint;
  end;
end;

procedure TDSTerminator.Paint;
begin
  inherited Paint;
  with ClientRect,Canvas do
  begin
    RoundRect(Left,Top,Right,Bottom,(Bottom-Top),(Bottom-Top));
    SetTextAlign(Handle, ta_Center or ta_Top);
    TextOut((Right-Left)div 2,3,IntToStr(FAmount));
    TextOut((Right-Left)div 2,(Bottom-Top)div 2 + 3,Caption);
  end;
end;

//=============== TDSAdvance ===============

constructor TDSAdvance.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Height:=40;
end;

function TDSAdvance.Enter(Transact: TTransact):boolean;
begin
  Result:=inherited Enter(Transact);
  if Result then
  begin
    if (FTimer=nil)then
      raise CreateError(7,Self);
    FTimer.Events.Add(GetInterval(Transact),Transact,Self);
    inc(FCurCount);
    Repaint;
  end;
end;

procedure TDSAdvance.Leave(Transact: TTransact);
begin
  dec(FCurCount);
  inherited Leave(Transact);
end;

procedure TDSAdvance.Reset;
begin
  FCurCount:=0;
  inherited Reset;
end;

procedure TDSAdvance.Paint;
begin
  inherited Paint;
  with Canvas,ClientRect do
  begin
    Rectangle(Left,Top,Right,Bottom);
    SetTextAlign(Handle, ta_Center or ta_Top);
    TextOut((Right-Left)div 2,3,Caption);
    SetTextAlign(Handle, TA_RIGHT or ta_Top);
    TextOut(Right - 5,(Bottom-Top)div 2+3,IntToStr(FAmount));
    SetTextAlign(Handle, TA_LEFT or ta_Top);
    TextOut(Left + 3 ,(Bottom-Top)div 2+3,IntToStr(FCurCount));
  end;
end;

//=============== TDSSplitter ===============

constructor TDSSplitter.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Height:=50;
  FSplitNumber := 3;
end;

function TDSSplitter.Enter(Transact: TTransact):boolean;
var i:integer;
    STransact:TTransact;
begin
  Result:=inherited Enter(Transact);
  if Result then
  begin
    for i := 2 to FSplitNumber do
    begin
      STransact:=TTransact.Create;
      STransact.Assign(Transact);
//      STransact.Parent:=Self;//!!!!
      try
        if Assigned(FOnSplit)then
          FOnSplit(Self,STransact,i);
        Leave(STransact);
      except
      end;
    end;
    Leave(Transact);
  end;
end;

procedure TDSSplitter.Paint;
var StepY:integer;
begin
  inherited Paint;
  with Canvas,ClientRect do
  begin
    StepY:=(Bottom-Top)div 3;

    Rectangle(Left,Top,Right,Bottom);
    Pen.Width:=1;
    PenPos:=Point(Left,Bottom);
    LineTo((Right-Left)div 2,Bottom-StepY);
    LineTo(Right,Bottom);

    SetTextAlign(Handle, ta_Center or ta_Top);
    TextOut((Right-Left)div 2,3,Caption);
    TextOut((Right-Left)div 2,StepY,IntToStr(FAmount));
    TextOut((Right-Left)div 2,Bottom-StepY+1,IntToStr(Number));
  end;
end;

procedure TDSSplitter.SetSplitNumber(const Value: integer);
begin
  FSplitNumber := Value;
  Invalidate;
end;

//=============== TDSAssembler ===============

procedure TDSAssembler.Clear;
begin
  inherited Clear;
  FAssemblingTransacts.Clear;
end;

constructor TDSAssembler.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Height:=50;
  FAssembleNumber := 3;
  FAssemblingTransacts:=TTransactList.Create;
end;

destructor TDSAssembler.Destroy;
begin
  FAssemblingTransacts.Free;
  inherited Destroy;
end;

function TDSAssembler.Enter(Transact: TTransact):boolean;
var CanAssm:boolean;
    TrList: TTransactList;
    i:integer;
begin
  Result:=inherited Enter(Transact);
  if Result then
  begin
    TrList:=TTransactList.Create;
    try
      i:=0;
      while(i<FAssemblingTransacts.Count)and(TrList.Count<FAssembleNumber-1)do
      begin
        CanAssm:=True;
        if Assigned(FOnAssembleQuery)then
          FOnAssembleQuery(Self,Transact,FAssemblingTransacts[i],CanAssm);
        if CanAssm then
          TrList.Add(FAssemblingTransacts[i]);
        inc(i);
      end;
      if TrList.Count=FAssembleNumber-1 then
      begin
        for i:=0 to TrList.Count-1 do
        begin
          FAssemblingTransacts.Remove(TrList[i]);
        end;
        Leave(Transact);
      end
      else
        FAssemblingTransacts.Add(Transact);
    finally
      TrList.Free;
    end;
  end;
end;

procedure TDSAssembler.Paint;
var StepY:integer;
begin
  inherited Paint;
  with Canvas,ClientRect do
  begin
    StepY:=(Bottom-Top)div 3;
    Rectangle(Left,Top,Right,Bottom);
    PenPos:=Point(Left,Top);
    Pen.Width:=1;
    LineTo((Right-Left)div 2,Top+StepY);
    LineTo(Right,Top);
    SetTextAlign(Handle, ta_Center or ta_Bottom);
    TextOut((Right-Left)div 2,Bottom-3,Caption);
    TextOut((Right-Left)div 2,Bottom-StepY,IntToStr(FAmount));
    TextOut((Right-Left)div 2,StepY-1,IntToStr(Number));
  end;
end;

procedure TDSAssembler.SetAssembleNumber(const Value: integer);
begin
  FAssembleNumber := Value;
  Invalidate;
end;

//===============   ===============

function IRandom(Min,Max:TDSTime):TDSTime;
begin
  Result:=Min+Random(Max-Min+1);
end;

function NegExp(AverageTime:TDSTime):TDSTime;
begin
  Result:=trunc(-AverageTime * ln(Random(High(AverageTime))/High(AverageTime)))
end;

function EmpDistribution(Probabls: array of real):integer;
var
 i:integer;
 Total,rnd :real;
begin
 rnd:=random;
 Total:=0;
 i:=0;
 repeat
   Total:=Total+Probabls[i];
   inc(i);
 until (i>High(Probabls))or(rnd<Total);
 Result:=i-1;
end;

//===============   ===============

function CreateError(const ErrCode: integer; const Device: TDSDevice = nil):EDSError;
var Msg:string;
begin
  case ErrCode of
    1: Msg:='   %s      -  !';
    2: Msg:='      !';
    3: Msg:=' : %s     !';
    4: Msg:='     !';
    5: Msg:=' : %s   -!';
    6: Msg:=' %s      !';
    7: Msg:='   Timer   %s!';
    8: Msg:='      %s!';
    9: Msg:='   %s!   .';
    10:Msg:=' %s     !';
    11:Msg:='     [0,1]';
    12:Msg:='        %s!';
    13:Msg:='  ,    %s!';
  end;
  Result:=EDSError.CreateFmt(Msg,[Device.Name]);
end;

procedure DrawBusyBar(const Canvas: TCanvas; const DrawRect: TRect; AMax,ACur:integer);
var NewRight:integer;
begin
  with Canvas, DrawRect do
  begin
    Pen.Width:=1;
    SetTextAlign(Handle, ta_Center or ta_Bottom);
    TextOut((Right-Left)div 2,Top-2,IntToStr(ACur));

    SetTextAlign(Handle, ta_Right or ta_Bottom);
    TextOut(Right,Top-2,IntToStr(AMax));

    Brush.Color:=clWhite;
    Rectangle(Left,Top,Right,Bottom);
    Brush.Color:=clNavy;
    FrameRect(DrawRect);
    if AMax<>0 then
    begin
      NewRight:=Left+Round((Right-Left)*(ACur/AMax));
      Rectangle(Left,Top,NewRight,Bottom);
    end;
  end;
end;
{
procedure DrawBusyBar(const Canvas: TCanvas; const DrawRect: TRect; AMax,ACur:integer);
var NewRight:integer;
    RectBMP: Graphics.TBitmap;
begin
  RectBMP := Graphics.TBitmap.Create;
  try
    RectBMP.Width:=DrawRect.Right-DrawRect.Left-2;
    RectBMP.Height:=DrawRect.Bottom-DrawRect.Top-2;
    with RectBMP do
    begin
      if AMax<>0 then
      begin
        Canvas.Brush.Color:=clNavy;
        NewRight:=Round(Width*(ACur/AMax));
        Canvas.Rectangle(0,0,NewRight,Height);
      end;
    end;
    with Canvas, DrawRect do
    begin
      Pen.Width:=1;
      Pen.Color:=clBlack;
      Brush.Color:=clWhite;
//      Rectangle(Left,Top,Right,Bottom);
//      FrameRect(DrawRect);

      SetTextAlign(Handle, ta_Center);
      TextRect(DrawRect,(Right-Left)div 2,Top+2,IntToStr(ACur));

      SetTextAlign(Canvas.Handle, ta_Right);
      TextOut(Right-2,Top+2,IntToStr(AMax));

      CopyMode := cmSrcInvert;
      Draw(Left+1, Top+1, RectBMP);
    end;
  finally
    RectBMP.Free;
  end;
end;
}
procedure DrawRomb(const Canvas: TCanvas; const DrawRect: TRect);
var CenterX,CenterY:integer;
begin
  with Canvas,DrawRect do
  begin
    CenterX:=(Right-Left)div 2;
    CenterY:=(Bottom-Top)div 2;
    Polygon([Point(CenterX,Top),Point(Right,CenterY),Point(CenterX,Bottom),Point(Left,CenterY)]);
  end;
end;

function TimeEvent(ATime:TDSTime;ADevice:TDSDevice;Transact:TTransact;AMethod:TTimeEventMethod;Id:integer):TTimeEvent;
begin
  Result.Time:=ATime;
  Result.Transact:=Transact;
  Result.Device:=ADevice;
  Result.Method:=AMethod;
  Result.Id:=Id;
end;

//===============  ===============

procedure Register;
begin
  RegisterComponents('Simulation',[TDSTimer,TDSGenerator,TDSOneChannelDevice,TDSMultiChannelDevice,TDSQueueOneChannelDevice,TDSQueueMultiChannelDevice,TDSTransfer,TDSTerminator,TDSAdvance,TDSQueue,TDSSplitter,TDSAssembler]);
end;

//===============  ===============

function TDSQueuelessChannelDevice.Enter(Transact: TTransact): boolean;
begin
  Result:=CanReserve(Transact)and inherited Enter(Transact);
end;

procedure TDSChannelDevice.SetPower(const Value: integer);
begin
  if FPower <> Value then
  begin
    FPower := Value;
    Invalidate;
  end;
end;

function TDSTransfer.GetCount(Index: integer): integer;
begin
  Result:=Counts[Index];
end;

initialization
  Randomize;
end.
